import click
import pandas as pd
import requests
import json
import os
import sys
from openpyxl import Workbook
from carpalrouting.grouping import LocationGrouping, DriverGrouping
from carpalrouting.optimize import Routing
from carpalrouting.models import Driver, Location, Coordinates, GeoArea
from decimal import Decimal


@click.command()
@click.option('--file', help='The absolute path of setting.json file')
def cli(file):
    if not file:
        print('Error. No schema file provided.')
    else:
        try:
            with open(file) as setting_file:
                print('Loading setting file...')
                settings = json.load(setting_file)
                print('Setting file loaded.')

                print('Reading driver data...')
                drivers = settings.get('drivers')
                print('Driver data loaded.')

                print('Reading location data...')
                df = pd.read_excel(settings.get('inputFile'))
                # df['sort'] = df['delivery_time_window'].str.extract('(\d+)',
                #                                                     expand=False).astype(int)
                # df.sort_values('sort', inplace=True, ascending=True)
                # df = df.drop('sort', axis=1)
                print('Location data loaded.')

                locations = {}
                for _, row in df.iterrows():
                    pickup_address = row['pickup_location_address']
                    pickup_window = tuple(
                        map_time_to_seconds(row['pickup_time_window']))

                    if pickup_address in locations:
                        new_item = {'address': row['delivery_address'],
                                    'coordinates':
                                        map_address_to_coordinates(
                                            row['delivery_address'],
                                            settings.get('googleAPIKey'))
                                        if settings.get('addressType') == 'address' else
                                        map_str_to_coordinates(
                                            row['delivery_address']),
                                    'delivery_window':
                                        map_time_to_seconds(
                                            row['delivery_time_window']),
                                    'capacity': row['capacity'],
                                    'order_time': 0,
                                    'ref': str(row['customer_order_number'])}
                        if pickup_window in locations[pickup_address]:
                            locations[pickup_address]
                            [pickup_window].append(new_item)
                        else:
                            locations[pickup_address][pickup_window] = [{
                                'address': pickup_address,
                                'coordinates':
                                    map_address_to_coordinates(
                                        pickup_address,
                                        settings.get('googleAPIKey'))
                                    if settings.get('addressType') == 'address'
                                    else map_str_to_coordinates(
                                        pickup_address),
                                'delivery_window':
                                    map_time_to_seconds(
                                        row['pickup_time_window']),
                                'capacity': 0,
                                'order_time': 0
                            }]
                            locations[pickup_address]
                            [pickup_window].append(new_item)
                    else:
                        locations[pickup_address] = {}

                        locations[pickup_address][pickup_window] = [{
                            'address': pickup_address,
                            'coordinates':
                                map_address_to_coordinates(
                                    pickup_address,
                                    settings.get('googleAPIKey'))
                                if settings.get('addressType') == 'address'
                                else map_str_to_coordinates(pickup_address),
                            'delivery_window':
                                map_time_to_seconds(row['pickup_time_window']),
                            'capacity': 0,
                            'order_time': 0
                        }]

                        locations[pickup_address][pickup_window].append({
                            'address': row['delivery_address'],
                            'coordinates':
                                map_address_to_coordinates(
                                    row['delivery_address'],
                                    settings.get('googleAPIKey'))
                                if settings.get('addressType') == 'address'
                                else map_str_to_coordinates(
                                    row['delivery_address']),
                            'delivery_window':
                                map_time_to_seconds(
                                    row['delivery_time_window']),
                            'capacity': row['capacity'],
                            'order_time': 0,
                            'ref': str(row['customer_order_number'])
                        })

                schedules = DriverGrouping().group_drivers_by_time_slot(
                    drivers=[Driver(id=item.get('id'),
                                    time_slots=item.get('time_slots'),
                                    capacity=item.get('capacity'),
                                    speed=item.get('speed')) for
                             item in drivers])

                data, not_found = LocationGrouping().group_locations(
                    locations,
                    schedules,
                    settings.get('mergeIdentical'))

                for key, val in data.items():
                    for k, v in val.items():
                        if len(v.get('locations')) < 2:
                            data[key][k]['locations'] = []

                print('Start route optimization...')
                book = Workbook()
                sheet = book.active

                inhouse = []
                public = []
                for pickup_address, pickup_windows in data.items():
                    for pickup_window, obj in pickup_windows.items():
                        locs = obj.get('locations')
                        driver_ids = obj.get('driver_ids')
                        while locs and driver_ids:
                            driver = None
                            for item in drivers:
                                if item.get('id') == driver_ids[0]:
                                    driver = item
                                    break

                            # find the indices for all filtered
                            # locations in data object
                            filtered_indices = []
                            for k, v in enumerate(locs):
                                if v.capacity <= driver.get('capacity'):
                                    filtered_indices.append(k)

                            indices = Routing(
                                locations=[loc for loc in locs if
                                           loc.capacity <=
                                           driver.get('capacity')],
                                num_vehicles=len(locs)).generate_routes(
                                    vehicle_capacity=driver.get('capacity'),
                                    speed=driver.get('speed'),
                                    service_time_unit=settings['serviceTime'])

                            if len(indices) == 0:                            
                                break
                            else:
                                # remove driver from available driver list
                                # only if a route has been built
                                driver_id = driver_ids.pop(0)

                                routes = []
                                for item in indices[0]:
                                    if ';' in locs[filtered_indices[item]].ref:
                                        for sub_item in locs[filtered_indices[item]].ref.split(';'):
                                            itm_info = sub_item.split(',')
                                            routes.append(Location(
                                                coordinates=Coordinates(locs[filtered_indices[item]].coordinates[0], 
                                                                        locs[filtered_indices[item]].coordinates[1]),
                                                address=locs[filtered_indices[item]].address,
                                                delivery_window=locs[filtered_indices[item]].delivery_window,
                                                capacity=float(itm_info[1]),
                                                ref=itm_info[0]))
                                    else:
                                        routes.append(locs[filtered_indices[item]])

                                    if filtered_indices[item] != 0:
                                        locs[filtered_indices[item]] = None
                                inhouse.append({driver_id: routes})
                                locs = list(filter(lambda x: x is not None, locs))
                                #sheet.append(tuple([str(item.to_json()) for item in routes]))

                        obj['locations'] = locs if len(locs) > 1 else [] 
                book.save(settings.get('outputFile'))

                # #There might be drivers and locations left over due to the merge identical, check again
                # for pickup_address, pickup_windows in data.items():
                #     for pickup_window, obj in pickup_windows.items():
                #         locs = obj.get('locations')

                #         #we have a merged location with capacity 20 here, thus the driver ID 7 is left over
                #         print([item.to_json() for item in locs])
                #         print(obj.get('driver_ids'))

                if settings['publicDriver']['required']:
                    print('Using public drivers...')
                    for pickup_address, pickup_windows in data.items():
                        for pickup_window, obj in pickup_windows.items():
                            locs = obj.get('locations')

                            if locs:
                                indices = Routing(
                                    locations=locs,
                                    num_vehicles=len(locs)).generate_routes(
                                        vehicle_capacity=settings['publicDriver']['specification']['capacity'],
                                        speed=settings['publicDriver']['specification']['speed'],
                                        service_time_unit=settings['serviceTime'])
                                if len(indices) == 0:
                                    break
                                else:
                                    for idx_list in indices:
                                        routes = []

                                        for item in idx_list:
                                            if ';' in locs[item].ref:
                                                for sub_item in locs[item].ref.split(';'):
                                                    itm_info = sub_item.split(',')
                                                    routes.append(Location(coordinates=Coordinates(locs[item].coordinates[0],
                                                                                                   locs[item].coordinates[1]),
                                                                           address=locs[item].address,
                                                                           delivery_window=locs[item].delivery_window,
                                                                           capacity=float(itm_info[1]),
                                                                           ref=itm_info[0]))
                                            else:
                                                routes.append(locs[item])
                                            if item != 0:
                                                locs[item] = None

                                        #sheet.append(tuple([str(item.to_json()) for item in routes]))
                                        public.append(routes)
                                    locs = list(filter(lambda x: x is not None, locs))
                                    obj['locations'] = locs if len(locs) > 1 else []

                Routing.merge_routes(inhouse_routes=inhouse,
                                     public_routes=public,
                                     drivers=drivers)

                sheet.append(tuple(['In-House drivers']))
                for item in inhouse:
                    for k, v in item.items():
                        sheet.append(tuple(['Driver ID: ' + str(k)]))
                        sheet.append(tuple([str(item.to_json()) for item in v]))
                
                sheet.append(tuple(['']))
                sheet.append(tuple(['Public drivers']))
                for route in public:
                    sheet.append(tuple([str(item.to_json()) for item in route]))                    

                sheet.append(tuple(['']))
                sheet.append(tuple(['Unmatched addresses']))
                sheet.append(tuple([item for item in not_found]))

                sheet.append(tuple(['']))
                sheet.append(tuple(['Unutilized drivers']))
                for pickup_address, pickup_windows in data.items():
                    for pickup_window, obj in pickup_windows.items():
                        for item in drivers:
                            if item.get('id') in obj.get('driver_ids'):
                                item['time_slots'] = [[map_seconds_to_time(ts[0]),
                                                       map_seconds_to_time(ts[1])] for ts in item['time_slots']]                             
                                sheet.append(tuple([json.dumps(item)]))

                book.save(settings.get('outputFile'))
                print('Done routing optimization.')
        except Exception as e:
            import traceback
            traceback.print_exc()
            print('Error. Unable to load the setting file.')
            sys.exit()


def map_address_to_coordinates(address, key):
    url = 'https://maps.googleapis.com/maps/api/geocode/json?address={}&key={}'.format(
        address.replace('\n', ' '), key)
    response = requests.get(url)
    content = json.loads(response.__dict__['_content'])
    if content.get('results'):
        loc = content['results'][0]['geometry']['location']
        return [loc.get('lat'), loc.get('lng')]
    else:
        return [0, 0]


def map_str_to_coordinates(geocode):    
    tmp = geocode.split(',')
    return [float(tmp[0]), float(tmp[1])]


def map_time_to_seconds(time):
    window = time.split('-')
    start = window[0].split(':')
    end = window[1].split(':')
    return [int(start[0])*3600+int(start[1])*60, int(end[0])*3600+int(end[1])*60]


def map_seconds_to_time(seconds):
    m, s = divmod(seconds, 60)
    h, m = divmod(m, 60)
    return '%02d:%02d' % (h, m) 


if __name__ == '__main__':
    cli()



import copy
from shapely.geometry import Point
from shapely.geometry.polygon import Polygon
from .models import Driver, Location, Wave, Coordinates, GeoArea


class DriverGrouping(object):
    def __init__(self):
        pass

    def group(self, drivers, locations,
              start_buffer_time=0, end_buffer_time=0):
        # Verifying data type
        illegal_driver_data = \
            [item for item in drivers if not isinstance(item, Driver)]
        illegal_location_data = \
            [item for item in locations if not isinstance(item, Location)]

        if illegal_driver_data:
            raise TypeError(
                'All element in the list should have the same type Driver.')

        if illegal_location_data:
            raise TypeError(
                'All element in the list should have the same type Location.')

        result = {}

        for driver in drivers:
            result[driver.id] = {}
            for time_slot in driver.time_slots:
                result[driver.id][(time_slot[0], time_slot[1])] = []
                for idx, val in enumerate(locations):
                    if not val:
                        continue
                    elif (time_slot[0] + start_buffer_time) < val.delivery_window[0] and \
                        val.delivery_window[1]< (time_slot[1] - end_buffer_time) and \
                        self.if_loc_capacity_within(val.capacity, driver.capacity):
                        result[driver.id][(time_slot[0], time_slot[1])].append(val)
                        locations[idx] = None

        result['unprocessed'] = \
            [item for item in locations if item is not None]
        return result

    def group_drivers_by_time_slot(self, drivers):
        # Verifying data type
        illegal_driver_data = \
            [item for item in drivers if not isinstance(item, Driver)]

        if illegal_driver_data:
            raise TypeError(
                'All element in the list should have the same type Driver.')

        time_slots = {}

        for driver in drivers:
            for time_slot in driver.time_slots:
                if (time_slot[0], time_slot[1]) in time_slots:
                    if time_slots[(time_slot[0], time_slot[1])]:
                        time_slots[(time_slot[0], time_slot[1])].append(driver)
                    else:
                        time_slots[(time_slot[0], time_slot[1])] = [driver]
                else:
                    time_slots[(time_slot[0], time_slot[1])] = [driver]
        return time_slots

    def if_loc_capacity_within(self, location_weight, driver_capacity):
        return driver_capacity > location_weight


class LocationGrouping(object):
    def __init__(self):
        pass

    def _group_locations_by_time_slot(self, locations,
                                      time_slot, threshhold=0):
        """
        Return locations that not included in the given time slot
        """
        result = [[], []]
        for loc in locations:
            if threshhold == 0:
                if time_slot[0] < loc.delivery_window[0] and \
                   loc.delivery_window[1] < time_slot[1]:
                    result[0].append(loc)
                else:
                    result[1].append(loc)
            else:
                # TODO: make use of threshhold argument
                if (time_slot[0] < loc.delivery_window[0] < time_slot[1] and
                    loc.delivery_window[0] == (time_slot[0]+time_slot[1])/2) or \
                   (time_slot[0] < loc.delivery_window[1] < time_slot[1] and
                   loc.delivery_window[0] == (time_slot[0]+time_slot[1])/2):
                    result[0].append(loc)
                else:
                    result[1].append(loc)
        return result

    def group_locations(self, locations, schedules, merge_identical=False):
        """
        This method requires the locations to be grouped by pickup locations

        Sample processed:
        {'pickup1': {(12000, 18000) : {
        (43140, 79200): [<carpalrouting.models.Location object at 0x110b62be0>],
        (0, 43140): [<carpalrouting.models.Location object at 0x1127d66d8>,
                        <carpalrouting.models.Location object at 0x112a15748>]}},
        'pickup2': {(12000, 18000) : {
        (28800, 79200): [<carpalrouting.models.Location object at 0x110778320>,
                         <carpalrouting.models.Location object at 0x112f0ecf8>,
                         <carpalrouting.models.Location object at 0x112f0ee48>]}},
        'pickup3': {(12000, 18000) : {
        (28800, 79200): [<carpalrouting.models.Location object at 0x112a15dd8>]}}}

        Sample unprocessed:
        {'pickup1': [],
         'pickup2': [],
         'pickup3': [<carpalrouting.models.Location object at 0x110778390>]}
        """
        result = {}
        not_found = set()
        copied_schedules = copy.deepcopy(schedules)

        if merge_identical:
            for pickup, pickup_windows in locations.items():
                for pickup_window, dropoff_list in pickup_windows.items():
                    if len(dropoff_list) > 2:
                        for k, v in enumerate(dropoff_list):
                            if k == 0 or not v:
                                continue
                            elif v.get('coordinates') == [0, 0]:
                                not_found.add(v.get('address'))
                                dropoff_list[k] = None
                                continue

                            for idx in range(k+1, len(dropoff_list)):
                                if dropoff_list[idx] and \
                                   dropoff_list[idx]['coordinates'] == [0, 0]:
                                        not_found.add(
                                            dropoff_list[idx]['address'])
                                        dropoff_list[idx] = None
                                elif dropoff_list[idx] and v['coordinates'] == \
                                    dropoff_list[idx]['coordinates'] and \
                                    v['delivery_window'] == dropoff_list[idx]['delivery_window']:
                                        if ',' not in v['ref']:
                                            v['ref'] = ','.join([v['ref'],
                                                                 str(v['capacity'])]) + \
                                                       ';' + ','.join([dropoff_list[idx]['ref'],
                                                                       str(dropoff_list[idx]['capacity'])])
                                        else:
                                            v['ref'] = v['ref'] + ';' + \
                                                       ','.join([dropoff_list[idx]['ref'],
                                                                 str(dropoff_list[idx]['capacity'])])

                                        v['capacity'] += dropoff_list[idx]['capacity']
                                        dropoff_list[idx] = None

                    elif dropoff_list[1].get('coordinates') == [0, 0]:
                        not_found.add(dropoff_list[1]['address'])
                        dropoff_list[1] = None

                    pickup_windows[pickup_window] = \
                        [loc for loc in dropoff_list if loc]

        for pickup, pickup_windows in locations.items():
            result[pickup] = {}
            for pickup_window, dropoff_list in pickup_windows.items():
                result[pickup][pickup_window] = {'locations': [Location(
                    coordinates=Coordinates(item.get('coordinates')[0],
                                            item.get('coordinates')[1]),
                    address=item.get('address'),
                    delivery_window=item.get('delivery_window'),
                    capacity=item.get('capacity', 1),
                    ref=item.get('ref', '')) for item in dropoff_list],
                                                 'driver_ids': []}

                total_capacity = 0
                min_pickup_time = 99999
                max_pickup_time = 0

                for item in result[pickup][pickup_window]['locations']:
                    total_capacity += item.capacity
                    if item.delivery_window[0] < min_pickup_time:
                        min_pickup_time = item.delivery_window[0]

                    if item.delivery_window[1] > max_pickup_time:
                        max_pickup_time = item.delivery_window[1]

                id_containers = result[pickup][pickup_window]['driver_ids']
                for time_slot, driver_list in copied_schedules.items():
                    if driver_list and time_slot[0] <= min_pickup_time and \
                       time_slot[1] >= max_pickup_time and total_capacity > 0:

                        if total_capacity >= driver_list[0].capacity:
                            while total_capacity > 0 and driver_list:
                                driver = driver_list.pop()
                                id_containers.append(driver.id)
                                total_capacity -= driver.capacity
                        else:
                            driver = driver_list.pop()
                            id_containers.append(driver.id)
                            total_capacity = 0

                        copied_schedules[time_slot] = driver_list

                    if total_capacity <= 0:
                        break

        for pickup, pickup_windows in locations.items():
            for pickup_window, dropoff_list in pickup_windows.items():
                for time_slot, driver_list in copied_schedules.items():
                    min_pickup_time = 99999
                    max_pickup_time = 0

                    for item in result[pickup][pickup_window]['locations']:
                        if item.delivery_window[0] < min_pickup_time:
                            min_pickup_time = item.delivery_window[0]

                        if item.delivery_window[1] > max_pickup_time:
                            max_pickup_time = item.delivery_window[1]

                    if driver_list and time_slot[0] <= min_pickup_time and \
                       time_slot[1] >= max_pickup_time:
                        result[pickup][pickup_window]['driver_ids'].extend(
                            [driver.id for driver in driver_list])
                        copied_schedules[time_slot] = []

        return result, not_found


class WaveGrouping(object):
    def __init__(self):
        pass

    def group(self, waves, locations, buffer_time=0):
        illegal_wave_data = \
            [item for item in waves if not isinstance(item, Wave)]
        illegal_location_data = \
            [item for item in locations if not isinstance(item, Location)]

        if illegal_wave_data:
            raise TypeError(
                'All element in the list should have the same type Wave.')

        if illegal_location_data:
            raise TypeError(
                'All element in the list should have the same type Location.')

        result = {}
        for wave in waves:
            result[wave.cutoff_time] = [locations.pop(idx) for idx, val in
                                        enumerate(locations) if
                                        val.order_time + buffer_time <
                                        wave.cutoff_time]

        if len(locations) > 0:
            result['unprocessed'] = locations

        return result



import copy
from shapely.geometry import Point
from shapely.geometry.polygon import Polygon
from .models import Coordinates, GeoArea, Location

class GeoTools(object):
    @staticmethod
    def is_point_in_area(edge_locations, locations):
        polygon = []
        for item in edge_locations:
            polygon.append((item[0], item[1]))

        return [loc for loc in locations if Polygon(polygon).contains(Point(loc[0], loc[1]))]


# import os, sys
# current_path = os.path.dirname(os.path.abspath(__file__))
# sys.path.insert(0, current_path + '/../')
# from carpalrouting.models import Driver, Location, Coordinates
# from carpalrouting.grouping import DriverGrouping, LocationGrouping
# from carpalrouting.optimize import Routing
# from decimal import Decimal
# import unittest

# class TestDriverLocationGrouping(unittest.TestCase):
#     def setUp(self):
#         self._drivers = [
#             {
#                 'id': 981,
#                 'time_slots': [[53000, 61600],[30000, 42000]],
#                 'capacity': 100,
#                 'speed': 20
#             },{
#                 'id': 1000,
#                 'time_slots': [[53000, 61600],[30000, 42000]],
#                 'capacity': 16,
#                 'speed': 5
#             }]
            
#         self._locations = [
#             {
#                 'coordinates': [-33.8815474, 150.9625217],
#                 'delivery_window': [64800, 72000],
#                 'capacity': 7.8,
#                 'order_time': 30660
#             },{
#                 'coordinates': [-33.8322542, 151.0548105],
#                 'delivery_window': [32400, 39600],
#                 'capacity': 15.8,
#                 'order_time': 33600
#             },{
#                 'coordinates': [-33.881716, 150.9136746],
#                 'delivery_window': [32400, 39600],
#                 'capacity': 3.8,
#                 'order_time': 45000
#             },{
#                 'coordinates': [-33.9545249, 150.9406193],
#                 'delivery_window': [54000, 61200],
#                 'capacity': 35.5,
#                 'order_time': 55800
#             }]

#     def test_driver_schedule_grouping(self):
#         self.assertRaises(TypeError, lambda: DriverGrouping().group(drivers=self._drivers,
#                                                                     locations=self._locations))

#         result = DriverGrouping().group(drivers=[Driver(id=item.get('id'),
#                                                         time_slots=item.get('time_slots'),
#                                                         capacity=item.get('capacity'),
#                                                         speed=item.get('speed')) for item in self._drivers],
#                                         locations=[Location(coordinates=Coordinates(item.get('coordinates')[0], item.get('coordinates')[1]),
#                                                             address=item.get('address'),
#                                                             delivery_window=item.get('delivery_window'),
#                                                             capacity=item.get('capacity'),
#                                                             order_time=item.get('order_time')) for item in self._locations])
        
#         self.assertEqual(len(list(result.items())), 3)

#     def test_driver_schedule_multiple_pickup_windows_grouping(self):
#         drivers = [{
#                 'id':1,
#                 'name': 'Driver 1',
#                 'time_slots': [[0, 43140], [43140, 79200]],
#                 'capacity': 500,
#                 'speed': 50
#             },
#             {
#                 'id':2,
#                 'name': 'Driver 2',
#                 'time_slots': [[0, 43140]],
#                 'capacity': 400,
#                 'speed': 50
#             },
#             {'id':3,
#                 'name': 'Driver 3',
#                 'time_slots': [[0, 43140]],
#                 'capacity': 500,
#                 'speed': 50
#             },
#             {'id':4,
#                 'name': 'Driver 4',
#                 'time_slots': [[28800, 43140]],
#                 'capacity': 500,
#                 'speed': 50
#             },
#             {'id':5,
#                 'name': 'Driver 5',
#                 'time_slots': [[28800, 79200]],
#                 'capacity': 500,
#                 'speed': 50
#             },
#             {'id':6,
#                 'name': 'Driver 6',
#                 'time_slots': [[0, 43140]],
#                 'capacity': 500,
#                 'speed': 50
#             },
#             {'id':7,
#                 'name': 'Driver 7',
#                 'time_slots': [[0, 43140], [58800, 80000]],
#                 'capacity': 500,
#                 'speed': 50
#             },
#             {'id':8,
#                 'name': 'Driver 8',
#                 'time_slots': [[28800, 79200]],
#                 'capacity': 500,
#                 'speed': 50
#             },
#             {'id':9,
#                 'name': 'Driver 9',
#                 'time_slots': [[0, 79200]],
#                 'capacity': 500,
#                 'speed': 50
#             },
#             {'id':10,
#                 'name': 'Driver 10',
#                 'time_slots': [[0, 50800], [60400, 89200]],
#                 'capacity': 800,
#                 'speed': 60
#             }]
            
#         locations = {
#             (1.3001, 102.0001):
#             {
#                 (12000, 20000):[
#                     {
#                         'coordinates': [-33.8815474, 150.9625217],
#                         'address': 'pick up 1',
#                         'delivery_window': [12000, 20000],
#                         'capacity': 0,
#                         'order_time': 30660
#                     },{
#                         'coordinates': [-33.8322542, 151.0548105],
#                         'address': 'Delivery 1',
#                         'delivery_window': [16400, 39600],
#                         'capacity': 65.8,
#                         'order_time': 33600
#                     },{
#                         'coordinates': [-33.881716, 150.9136746],
#                         'address': 'Delivery 2',
#                         'delivery_window': [22400, 39600],
#                         'capacity': 43.8,
#                         'order_time': 45000
#                     }
#                 ],(52000, 58600):[
#                     {
#                         'coordinates': [-33.8815431, 151.9234],
#                         'address': 'pickup 2',
#                         'delivery_window': [54800, 58600],
#                         'capacity': 0,
#                         'order_time': 30660
#                     },{
#                         'coordinates': [-32.8311, 152.2341],
#                         'address': 'delivery 3',
#                         'delivery_window': [60000, 68000],
#                         'capacity': 15.8,
#                         'order_time': 33600
#                     }
#                 ]
#             },
#             (1.3002, 102.0002):
#             {
#                 (40000, 44000):[{
#                     'coordinates': [-33.9545249, 150.9406193],
#                     'address': 'pickup 3',
#                     'delivery_window': [40000, 44000],
#                     'capacity': 0,
#                     'order_time': 55800
#                 },{
#                     'coordinates': [-33.9375904, 150.8350134],
#                     'address': 'delivery 4',
#                     'delivery_window': [44000, 61200],
#                     'capacity': 15.5,
#                     'order_time': 35800
#                 },{
#                     'coordinates': [-33.8800401, 151.1104426],
#                     'address': 'delivery 5',
#                     'delivery_window': [44000, 61200],
#                     'capacity': 25,
#                     'order_time': 45800
#                 }]
#             },
#             (1.3003, 102.0003):
#                 {(20000, 26800):[{
#                     'coordinates': [-33.7171208, 151.0952565],
#                     'address': 'pickup 4',
#                     'delivery_window': [20000, 26800],
#                     'capacity': 0,
#                     'order_time': 55800
#                 },{
#                     'coordinates': [-33.9545249, 150.9406193],
#                     'address': 'delivery 6',
#                     'delivery_window': [36000, 41200],
#                     'capacity': 35.5,
#                     'order_time': 55800
#                 }]
#             }
#         }

#         schedules = DriverGrouping().group_drivers_by_time_slot(drivers=[Driver(id=item.get('id'),
#                                                                                 time_slots=item.get('time_slots'),
#                                                                                 capacity=item.get('capacity'),
#                                                                                 speed=item.get('speed')) for item in drivers])

#         data = LocationGrouping().group_locations(locations, schedules)

#         #self.assertEqual(len(all_routes), 4)



import os
import sys
import time
current_path = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, current_path + '/../')
import unittest
from decimal import Decimal
from carpalrouting.optimize import Routing
from carpalrouting.models import Driver, Location, Coordinates
from carpalrouting.grouping import DriverGrouping, LocationGrouping
from data import drivers, locations


class TestOptimizer(unittest.TestCase):   
    def test_optimization_with_predefined_data(self):
        
        schedules = DriverGrouping().group_drivers_by_time_slot(
            drivers=[Driver(id=item.get('id'),
                            time_slots=item.get('time_slots'),
                            capacity=item.get('capacity'),
                            speed=item.get('speed')) for item in drivers])

        data, _ = LocationGrouping().group_locations(locations, schedules, True)

        self.assertEqual(len(_), 0)

        inhouse = []
        for pickup_address, pickup_windows in data.items():
            for pickup_window, obj in pickup_windows.items():
                locs = obj.get('locations')
                driver_ids = obj.get('driver_ids')
                while locs and driver_ids:
                    driver = None 
                    for item in drivers:
                        if item.get('id') == driver_ids[0]:
                            driver = item
                            break

                    filtered_indices = []
                    for k, v in enumerate(locs):
                        if v.capacity <= driver.get('capacity'):
                            filtered_indices.append(k)
                            
                    indices = Routing(locations=[loc for loc in locs if loc.capacity <= driver.get('capacity')],
                                      num_vehicles=len(locs)).generate_routes(
                                          vehicle_capacity=driver.get('capacity'),
                                          speed=driver.get('speed'),
                                          service_time_unit=1800)

                    if len(indices) == 0:                       
                        break
                    else:
                        #remove driver from available driver list only if a route has been built
                        driver_id = driver_ids.pop(0)                            

                        routes = []
                        for item in indices[0]:
                            if ';' in locs[filtered_indices[item]].ref:
                                for sub_item in locs[filtered_indices[item]].ref.split(';'):
                                    itm_info = sub_item.split(',')
                                    routes.append(
                                        Location(coordinates = Coordinates(
                                                    locs[filtered_indices[item]].coordinates[0], 
                                                    locs[filtered_indices[item]].coordinates[1]),
                                                 address = locs[filtered_indices[item]].address,
                                                 delivery_window = locs[filtered_indices[item]].delivery_window,
                                                 capacity = float(itm_info[1]),
                                                 ref = itm_info[0]))
                            else:
                                routes.append(locs[filtered_indices[item]])
                            if filtered_indices[item] != 0:
                                locs[filtered_indices[item]] = None
                        inhouse.append({driver_id: routes})
                        locs = list(filter(lambda x: x is not None, locs))
                obj['locations'] = locs if len(locs) > 1 else []


        public = []
        for pickup_address, pickup_windows in data.items():
            for pickup_window, obj in pickup_windows.items():
                locs = obj.get('locations')
                if locs:
                    indices = Routing(
                                locations=locs,
                                num_vehicles=len(locs)).generate_routes(
                                                                    vehicle_capacity=30,
                                                                    speed=30,
                                                                    service_time_unit=1800)

                    if len(indices) == 0:                    
                        break
                    else:                                                
                        for idx_list in indices:
                            routes = []

                            for item in idx_list:
                                if ';' in locs[item].ref:
                                    for sub_item in locs[item].ref.split(';'):
                                        itm_info = sub_item.split(',')
                                        routes.append(
                                            Location(coordinates = Coordinates(locs[item].coordinates[0], 
                                                                               locs[item].coordinates[1]),
                                                     address = locs[item].address,
                                                     delivery_window = locs[item].delivery_window,
                                                     capacity = float(itm_info[1]),
                                                     ref = itm_info[0]))
                                else:
                                    routes.append(locs[item])

                                if item != 0:
                                    locs[item] = None
                            public.append(routes)
                        locs = list(filter(lambda x: x is not None, locs))
                        obj['locations'] = locs if len(locs) > 1 else []
        
        inhouse, public = Routing.merge_routes(inhouse_routes=inhouse,
                                               public_routes=public,
                                               drivers=drivers)
        

        self.assertEqual(len(inhouse), 5)
        self.assertEqual(len(public), 3)
